/* GIMP RGB C-Source image dump (gray2x3.c) */

static const struct {
  guint  	 width;
  guint  	 height;
  guint  	 bytes_per_pixel; /* 2:RGB16, 3:RGB, 4:RGBA */ 
  guint8 	 pixel_data[2 * 3 * 3 + 1];
} gimp_image = {
  2, 3, 3,
  "\0\0\0???\200\200\200\276\276\276\345\345\345\377\377\377",
};

