<style type="text/css">
  h5 {
  margin-left: 1%;
  }
  pre.screen {
  margin-left: 2%;
  margin-right: 2%;
  background-color:#E0E0E0;
  }
  
  div.variablelist {
  margin-left: 2%;
  }
  span.term {
  font-style: italic
  }
  p.sectiontoc {
      margin-top:3px;
      margin-bottom:3px;
      margin-left: 2%;
  }
  p.progdescr{
      margin-top:3px;
      margin-bottom:3px;
      margin-left: 5%;
  }
  p.plugdescr{
      margin-top:3px;
      margin-bottom:3px;
      margin-left: 5%;
  }

  div.cmdsynopsis {
  margin-left: 2%;
  }
  p.consumer {
      margin-left: 2%;
  }
  p.plugindescr {
      margin-left: 3%;
  }

  p.pluginsubdescr {
      margin-left: 6%;
  }


</style>

