

#include <mia/core/test_core.cc>
#include <mia/core/test_cost.cc>
#include <mia/core/test_filter.cc>
#include <mia/core/test_kernels.cc>
#include <mia/core/test_optparam.cc>
#include <mia/core/test_optionparser.cc>
#include <mia/core/test_parameter.cc>
#include <mia/core/test_pixeltype.cc>
#include <mia/core/test_register.cc>
#include <mia/core/test_sqmin.cc>
#include <mia/core/test_streamredir.cc>
#include <mia/core/test_dictmap.cc>
#include <mia/core/test_fifofilter.cc>
#include <mia/core/test_probmap.cc>
